/*
 * Symbolic model checking of CSL formulas.
 *
 * Copyright (C) 2002 Dave Parker
 * Copyright (C) 2003--2004 Carnegie Mellon University
 *
 * This file is part of Ymer.
 *
 * Ymer is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Ymer is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Ymer; if not, write to the Free Software Foundation,
 * Inc., #59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "formulas.h"

#include <float.h>

#include <cmath>
#include <iostream>
#include <optional>
#include <stdexcept>

#include "hybrid.h"
#include "src/compiled-property.h"
#include "src/ddmodel.h"

#include "cudd.h"

namespace {

class SymbolicVerifier final : public CompiledPropertyVisitor,
                               public CompiledPathPropertyVisitor {
 public:
  explicit SymbolicVerifier(const DecisionDiagramModel* dd_model,
                            bool top_level_property, double epsilon);

  BDD result() const { return result_; }

 private:
  void DoVisitCompiledNaryProperty(
      const CompiledNaryProperty& property) override;
  void DoVisitCompiledNotProperty(const CompiledNotProperty& property) override;
  void DoVisitCompiledProbabilityThresholdProperty(
      const CompiledProbabilityThresholdProperty& property) override;
  void DoVisitCompiledProbabilityEstimationProperty(
      const CompiledProbabilityEstimationProperty& property) override;
  void DoVisitCompiledExpressionProperty(
      const CompiledExpressionProperty& property) override;
  void DoVisitCompiledUntilProperty(
      const CompiledUntilProperty& path_property) override;

  const DecisionDiagramModel* dd_model_;
  bool estimate_;
  bool top_level_property_;
  double epsilon_;
  double threshold_;
  bool strict_;
  BDD result_;
};

SymbolicVerifier::SymbolicVerifier(const DecisionDiagramModel* dd_model,
                                   bool top_level_property, double epsilon)
    : dd_model_(dd_model),
      estimate_(false),
      top_level_property_(top_level_property),
      epsilon_(epsilon),
      result_(dd_model->manager().GetConstant(false)) {}

void SymbolicVerifier::DoVisitCompiledNaryProperty(
    const CompiledNaryProperty& property) {
  switch (property.op()) {
    case CompiledNaryOperator::AND: {
      BDD result = dd_model_->manager().GetConstant(true);
      if (property.has_expr_operand()) {
        result = BDD(property.expr_operand().expr().dd().value()) && result;
      }
      for (const CompiledProperty& operand : property.other_operands()) {
        operand.Accept(this);
        result = result_ && result;
      }
      result_ = result;
      break;
    }
    case CompiledNaryOperator::OR: {
      BDD result = dd_model_->manager().GetConstant(false);
      if (property.has_expr_operand()) {
        result = BDD(property.expr_operand().expr().dd().value()) || result;
      }
      for (const CompiledProperty& operand : property.other_operands()) {
        operand.Accept(this);
        result = result_ || result;
      }
      result_ = result;
      break;
    }
    case CompiledNaryOperator::IFF: {
      BDD result = dd_model_->manager().GetConstant(false);
      bool has_result = false;
      if (property.has_expr_operand()) {
        result = BDD(property.expr_operand().expr().dd().value());
        has_result = true;
      }
      for (const CompiledProperty& operand : property.other_operands()) {
        operand.Accept(this);
        result = has_result ? result == result_ : result_;
        has_result = true;
      }
      result_ = result;
      break;
    }
  }
  result_ = dd_model_->reachable_states() && result_;
}

void SymbolicVerifier::DoVisitCompiledNotProperty(
    const CompiledNotProperty& property) {
  property.operand().Accept(this);
  result_ = dd_model_->reachable_states() && !result_;
}

void SymbolicVerifier::DoVisitCompiledProbabilityThresholdProperty(
    const CompiledProbabilityThresholdProperty& property) {
  double threshold = property.threshold();
  bool estimate = false;
  bool strict = property.op() == CompiledProbabilityThresholdOperator::GREATER;
  std::swap(threshold_, threshold);
  std::swap(estimate_, estimate);
  std::swap(strict_, strict);
  property.path_property().Accept(this);
  std::swap(threshold_, threshold);
  std::swap(estimate_, estimate);
  std::swap(strict_, strict);
}

void SymbolicVerifier::DoVisitCompiledProbabilityEstimationProperty(
    const CompiledProbabilityEstimationProperty& property) {
  double threshold = 0.5;
  bool estimate = true;
  bool strict = true;
  std::swap(threshold_, threshold);
  std::swap(estimate_, estimate);
  std::swap(strict_, strict);
  property.path_property().Accept(this);
  std::swap(threshold_, threshold);
  std::swap(estimate_, estimate);
  std::swap(strict_, strict);
}

void SymbolicVerifier::DoVisitCompiledExpressionProperty(
    const CompiledExpressionProperty& property) {
  result_ = dd_model_->reachable_states() && BDD(property.expr().dd().value());
}

// Matrix vector multiplication stuff.
static void mult_rec(const std::vector<double>& soln, double unif,
                     HDDMatrix* hddm, HDDNode* hdd, int level, long row,
                     long col, std::vector<double>* soln2) {
  if (hdd == hddm->zero) {
    // It is the zero node.
    return;
  } else if (hdd->sb) {
    // There is a sparse bit.
    for (auto element : *hdd->sb) {
      (*soln2)[row + element.row()] +=
          soln[col + element.column()] * element.value() * unif;
    }
    return;
  } else if (level == hddm->num_levels) {
    // We have reached the bottom.
    (*soln2)[row] += soln[col] * hdd->type.val * unif;
    return;
  }
  HDDNode* e = hdd->type.kids.e;
  if (e != hddm->zero) {
    mult_rec(soln, unif, hddm, e->type.kids.e, level + 1, row, col, soln2);
    mult_rec(soln, unif, hddm, e->type.kids.t, level + 1, row, col + e->off,
             soln2);
  }
  HDDNode* t = hdd->type.kids.t;
  if (t != hddm->zero) {
    mult_rec(soln, unif, hddm, t->type.kids.e, level + 1, row + hdd->off, col,
             soln2);
    mult_rec(soln, unif, hddm, t->type.kids.t, level + 1, row + hdd->off,
             col + t->off, soln2);
  }
}

// Poisson tail bounds.
static double tail_bound(int n, double lambda, int left, int right,
                         double epsilon) {
  int m = (int)lambda;
  if (m >= 2 && n >= 2 + m) {
    double sqrt_lambda = sqrt(lambda);
    double sqrt_2pi = sqrt(2.0 * M_PI);
    double k = (n - m - 3.0 / 2.0) / M_SQRT2 / sqrt_lambda;
    double a = (1.0 - 1.0 / lambda) * exp(1.0 / 16.0) * M_SQRT2;
    double d =
        1.0 / (1.0 - exp(-2.0 / 9.0 * (k * M_SQRT2 * sqrt_lambda + 3.0 / 2.0)));
    double tail = a * d * exp(-k * k / 2.0) / k / sqrt_2pi;
    if (left > 0) {
      k = (m - n - 3.0 / 2.0) / sqrt_lambda;
      double b = (1.0 + 1.0 / lambda) * exp(1.0 / 8.0 / lambda);
      tail += b * exp(-k * k / 2.0) / k / sqrt_2pi;
    }
    return (tail < 1.0) ? tail : 1.0;
  } else {
    return 1.0;
  }
}

// Finds truncation points L and R and computes weights for a sum of
// Poisson probabilities with rate lambda given an error bound of
// epsilon (Fox & Glynn, CACM 31(4):440-445, 1988).
static void fox_glynn_weighter(int& left, int& right, double*& weights,
                               double& weight_sum, double lambda,
                               double epsilon) {
  // Set the mode.
  int m = int(lambda);

  // Make sure lambda is positive.
  if (lambda <= 0.0) {
    throw std::invalid_argument("lambda <= 0.0");
  }

  // Find truncation points L and R.
  double sqrt_lambda = sqrt(lambda);
  double sqrt_2pi = sqrt(2.0 * M_PI);
  if (lambda < 25.0) {
    left = 0;
    if (exp(-lambda) < DBL_MIN) {
      throw std::underflow_error("potential underflow for Fox-Glynn");
    }
  } else {  // lambda >= 25.0
    // Find left using Corollary 2 with actual lambda.
    double b = (1.0 + 1.0 / lambda) * exp(1.0 / 8.0 / lambda);
    int i = 2;
    for (; i < m; i++) {
      double k = (i - 3.0 / 2.0) / sqrt_lambda;
      double T = b * exp(-k * k / 2.0) / k / sqrt_2pi;
      if (T < epsilon / 2) {
        break;
      }
    }
    left = m - i;
  }
  if (lambda < 400.0) {
    // Find right using Corollary 1 with lambda=400.
    double a = (1.0 + 1.0 / 400.0) * exp(1.0 / 16.0) * M_SQRT2;
    int i = 2;
    int i_max = (400 + 3) / 2;
    for (; i < i_max; i++) {
      double k = (i - 3.0 / 2.0) / M_SQRT2 / 20.0;
      double d =
          1.0 / (1.0 - exp(-2.0 / 9.0 * (k * M_SQRT2 * 20.0 + 3.0 / 2.0)));
      double Q = a * d * exp(-k * k / 2.0) / k / sqrt_2pi;
      if ((left == 0 && Q < epsilon) || (left > 0 && Q < epsilon / 2)) {
        break;
      }
    }
    right = m + i;
  } else {  // lambda >= 400.0
    // Find right using Corollary 1 with actual lambda.
    double a = (1.0 + 1.0 / lambda) * exp(1.0 / 16.0) * M_SQRT2;
    int i = 2;
    int i_max = int((lambda + 3.0) / 2.0);
    for (; i < i_max; i++) {
      double k = (i - 3.0 / 2.0) / M_SQRT2 / sqrt_lambda;
      double d =
          1.0 /
          (1.0 - exp(-2.0 / 9.0 * (k * M_SQRT2 * sqrt_lambda + 3.0 / 2.0)));
      double Q = a * d * exp(-k * k / 2.0) / k / sqrt_2pi;
      if ((left == 0 && Q < epsilon) || (left > 0 && Q < epsilon / 2)) {
        break;
      }
    }
    right = m + i;
  }

  // Compute weights.
  weights = new double[right - left + 1];
  weights[m - left] = (DBL_MAX * 1e-10) / (right - left);
  // Down.
  for (int j = m; j > left; j--) {
    weights[j - 1 - left] = (j / lambda) * weights[j - left];
  }
  // Up.
  if (lambda < 400.0) {
    if (right > 600) {
      throw std::underflow_error("potential undeflow for Fox-Glynn");
    } else {
      for (int j = m; j < right;) {
        double q = lambda / (j + 1.0);
        if (weights[j - left] > DBL_MIN / q) {
          weights[j + 1 - left] = q * weights[j - left];
          j++;
        } else {
          right = j;
        }
      }
    }
  } else {
    for (int j = m; j < right; j++) {
      weights[j + 1 - left] = (lambda / (j + 1.0)) * weights[j - left];
    }
  }

  // Compute sum of weights.
  weight_sum = 0.0;
  int s = left;
  int t = right;
  while (s < t) {
    if (weights[s - left] <= weights[t - left]) {
      weight_sum += weights[s - left];
      s++;
    } else {
      weight_sum += weights[t - left];
      t--;
    }
  }
  weight_sum += weights[s - left];
}

void SymbolicVerifier::DoVisitCompiledUntilProperty(
    const CompiledUntilProperty& path_property) {
  /*
   * Detect trivial cases.
   */
  if (threshold_ == 0 && !strict_) {
    /* Satisfied by all reachable states. */
    result_ = dd_model_->reachable_states();
    return;
  } else if (threshold_ == 1 && strict_) {
    /* Not satisfied by any states. */
    result_ = dd_model_->manager().GetConstant(false);
    return;
  }

  /*
   * Verify postcondition property.
   */
  bool top_level_property = false;
  std::swap(top_level_property_, top_level_property);
  path_property.post_property().Accept(this);
  BDD dd2 = result_;
  std::swap(top_level_property_, top_level_property);
  if (path_property.max_time() == 0) {
    /* No time is allowed to pass so solution is simply dd2. */
    result_ = dd2;
    return;
  }

  /*
   * Verify precondition property.
   */
  std::swap(top_level_property_, top_level_property);
  path_property.pre_property().Accept(this);
  BDD dd1 = result_;
  std::swap(top_level_property_, top_level_property);

  if (path_property.min_time() > 0) {
    // TODO(hlsyounes): implement support for interval time bounds.
    throw std::invalid_argument("interval time bounds not supported");
  }

  /*
   * Compute BDD representing non-absorbing states in transformed model.
   */
  BDD maybe = dd1 && !dd2;
  if (maybe.is_same(dd_model_->manager().GetConstant(false))) {
    /* All states are absorbing so solution is simply dd2. */
    result_ = dd2;
    return;
  }

  /* ---------------------------------------------------------------------- */
  /* Transient analysis. */

  /* Time limit. */
  double time = path_property.max_time();
  /* Number of states. */
  size_t nstates = dd_model_->odd().state_count();

  /*
   * Build HDD for matrix.
   */
  if (top_level_property_) {
    std::cout << "Building hybrid MTBDD matrix...";
  }
  ADD ddR = dd_model_->rate_matrix() * ADD(maybe);
  HDDMatrix* hddm =
      build_hdd_matrix(dd_model_->manager(), ddR, dd_model_->odd().root());
  if (top_level_property_) {
    std::cout << hddm->num_nodes << " nodes, " << hddm->num_sb
              << " sparse matrices (" << hddm->sbl << " levels)." << std::endl;
  }

  /* Get vector of diagonals. */
  double* diags = hdd_negative_row_sums(hddm, nstates);

  /*
   * Find max diagonal element.
   */
  double max_diag = diags[0];
  for (size_t i = 1; i < nstates; i++) {
    if (diags[i] < max_diag) {
      max_diag = diags[i];
    }
  }
  max_diag = -max_diag;

  /* Uniformization constant. */
  double unif = 1.0 / (1.01 * max_diag);

  /*
   * Modify diagonals.
   */
  for (size_t i = 0; i < nstates; i++) {
    diags[i] = unif * diags[i] + 1;
  }

  /*
   * Create solution/iteration vectors.
   */
  std::vector<double> soln = dd_model_->odd().AddToVector(ADD(dd2));
  std::vector<double> soln2(nstates);
  std::optional<int> init;
  if (top_level_property_) {
    init = dd_model_->initial_state_index();
  }
  std::vector<double> sum(init.has_value() ? 1 : nstates, 0);

  /*
   * Compute poisson probabilities.
   */
  int left, right;
  double* weights;
  double weight_sum;
  if (top_level_property_) {
    std::cout << "Uniformization: " << max_diag << "*" << time << " = "
              << (max_diag * time) << std::endl;
  }
  fox_glynn_weighter(left, right, weights, weight_sum, 1.01 * max_diag * time,
                     epsilon_);
  if (top_level_property_) {
    std::cout << "Fox-Glynn: left = " << left << ", right = " << right
              << std::endl;
  }

  /*
   * Iterations before left bound to update vector.
   */
  if (top_level_property_) {
    std::cout << "Computing probabilities";
  }
  int iters;
  bool done = false;
  bool steady = false;
  for (iters = 1; iters < left && !done; iters++) {
    if (top_level_property_) {
      if (iters % 1000 == 0) {
        std::cout << ':';
      } else if (iters % 100 == 0) {
        std::cout << '.';
      }
    }
    /*
     * Matrix vector multiplication.
     */
    for (size_t i = 0; i < nstates; i++) {
      soln2[i] = diags[i] * soln[i];
    }
    mult_rec(soln, unif, hddm, hddm->top, 0, 0, 0, &soln2);
    /*
     * Check for steady state convergence.
     */
    if (!estimate_) {
      done = true;
      double sqnorm = 0.0;
      double sqbound = epsilon_ * epsilon_ / 64.0;
      for (size_t i = 0; i < nstates; i++) {
        double diff = soln2[i] - soln[i];
        sqnorm += diff * diff;
        if (sqnorm > sqbound) {
          done = false;
          break;
        }
      }
      if (done) {
        steady = true;
        if (init.has_value()) {
          sum[0] = soln2[init.value()];
        } else {
          for (size_t i = 0; i < nstates; i++) {
            sum[i] = soln2[i];
          }
        }
        break;
      }
    }
    /*
     * Prepare for next iteration.
     */
    soln.swap(soln2);
  }

  /*
   * Accumulate weights.
   */
  for (; iters <= right && !done; iters++) {
    if (top_level_property_) {
      if (iters % 1000 == 0) {
        std::cout << ':';
      } else if (iters % 100 == 0) {
        std::cout << '.';
      }
    }
    if (iters > 0) {
      /*
       * Matrix vector multiplication.
       */
      for (size_t i = 0; i < nstates; i++) {
        soln2[i] = diags[i] * soln[i];
      }
      mult_rec(soln, unif, hddm, hddm->top, 0, 0, 0, &soln2);
      /*
       * Check for steady state convergence.
       */
      if (!estimate_) {
        done = true;
        double sqnorm = 0.0;
        double sqbound = epsilon_ * epsilon_ / 64.0;
        for (size_t i = 0; i < nstates; i++) {
          double diff = soln2[i] - soln[i];
          sqnorm += diff * diff;
          if (sqnorm > sqbound) {
            done = false;
            break;
          }
        }
        if (done) {
          steady = true;
          double weight = 0.0;
          for (int i = iters; i <= right; i++) {
            weight += weights[i - left] / weight_sum;
          }
          if (init.has_value()) {
            sum[0] += weight * soln2[init.value()];
          } else {
            for (size_t i = 0; i < nstates; i++) {
              sum[i] += weight * soln2[i];
            }
          }
          break;
        }
      }
      /*
       * Prepare for next iteration.
       */
      soln.swap(soln2);
    }
    /*
     * Add to sum.
     */
    size_t num_pass = 0;
    for (size_t i = 0; i < nstates; i++) {
      if (init.has_value()) {
        sum[0] += weights[iters - left] / weight_sum * soln[init.value()];
      } else {
        sum[i] += weights[iters - left] / weight_sum * soln[i];
      }
      double slack =
          tail_bound(iters, 1.01 * max_diag * time, left, right, epsilon_);
      if (VLOG_IS_ON(2) && init.has_value()) {
        LOG(INFO) << iters << " p_init in [" << sum[0] << ','
                  << (sum[0] + slack) << "]"
                  << "; threshold = " << threshold_;
      }
      if (!estimate_) {
        bool pass = false;
        if (strict_) {
          pass = (sum[i] + slack <= threshold_ || sum[i] > threshold_);
        } else {
          pass = (sum[i] + slack < threshold_ || sum[i] >= threshold_);
        }
        if (pass) {
          num_pass++;
          if (init.has_value() || num_pass == nstates) {
            done = true;
            break;
          }
        }
      }
      if (init.has_value()) {
        break;
      }
    }
    if (done) {
      break;
    }
  }
  if (iters > right) {
    iters = right;
  }
  if (top_level_property_) {
    std::cout << ' ' << iters << " iterations." << std::endl;
    if (steady) {
      std::cout << "Steady state detected." << std::endl;
    }
    if (estimate_) {
      double min_value = sum[0];
      double max_value = sum[0];
      for (size_t i = 1; i < sum.size(); ++i) {
        min_value = std::min(min_value, sum[i]);
        max_value = std::max(max_value, sum[i]);
      }
      std::cout.precision(10);
      std::cout << "Pr[" << path_property.string() << "] ";
      if (min_value == max_value) {
        std::cout << "= " << min_value;
      } else {
        std::cout << "in [" << min_value << "," << max_value << "]";
      }
      std::cout << std::endl;
      std::cout.precision(6);
    }
  }
  /*
   * Free memory.
   */
  free_hdd_matrix(hddm);
  delete diags;
  delete weights;

  std::function<bool(double)> satisfies_threshold;
  if (strict_) {
    satisfies_threshold = [this](double value) { return value > threshold_; };
  } else {
    satisfies_threshold = [this](double value) { return value >= threshold_; };
  }
  if (init.has_value()) {
    if (satisfies_threshold(sum[0])) {
      result_ = dd_model_->initial_states();
    } else {
      result_ = dd_model_->manager().GetConstant(false);
    }
  } else {
    result_ = dd_model_->odd().VectorToBdd(dd_model_->manager(), sum,
                                           satisfies_threshold);
  }
}

}  // namespace

BDD Verify(const CompiledProperty& property,
           const DecisionDiagramModel& dd_model, bool top_level_property,
           double epsilon) {
  SymbolicVerifier verifier(&dd_model, top_level_property, epsilon);
  property.Accept(&verifier);
  return verifier.result();
}

BDD VerifyExistsUntil(const DecisionDiagramModel& dd_model, const BDD& pre,
                      const BDD& post) {
  const int var_count = dd_model.manager().GetVariableCount();
  std::vector<int> row_to_col(var_count);
  for (int i = 0; i < var_count; i += 2) {
    row_to_col[i] = row_to_col[i + 1] = i + 1;
  }
  const BDD col_cube = dd_model.manager().GetCube(
      dd_model.manager().GetBddVariableArray(1, 2, var_count + 1));
  const BDD trans = pre && BDD(dd_model.rate_matrix());
  BDD states = dd_model.manager().GetConstant(false);
  BDD next_states = post;
  do {
    states = next_states;
    next_states =
        states || trans.AndAbstract(states.Permutation(row_to_col), col_cube);
  } while (!states.is_same(next_states));
  return states;
}
